import typing as t

from getajob.vendor.kafka.repository import KafkaProducerRepository
from getajob.vendor.kafka.models import KafkaEventConfig, KafkaTopic, KafkaCompanyEnum
from getajob.abstractions.repository import ParentRepository, RepositoryDependencies
from getajob.abstractions.models import Entity, EntityModels, UserAndDatabaseConnection
from getajob.contexts.companies.details.models import SetCompanyDetails
from getajob.contexts.companies.details.repository import CompanyDetailsRepository

from .models import (
    ClerkCompanyWebhookEvent,
    ClerkCompany,
    ClerkCompanyDeleted,
    ClerkCompanyWebhookType,
)

entity_models = EntityModels(entity=ClerkCompany)


class WebhookCompanyRepository(ParentRepository):
    def __init__(
        self,
        request_scope: UserAndDatabaseConnection,
        kafka: t.Optional[KafkaProducerRepository] = None,
    ):
        kafka_event_config = KafkaEventConfig(
            topic=KafkaTopic.companies, message_type_enum=KafkaCompanyEnum
        )
        self.request_scope = request_scope
        super().__init__(
            RepositoryDependencies(
                user_id=request_scope.initiating_user_id,
                db=request_scope.db,
                collection_name=Entity.COMPANIES.value,
                entity_models=entity_models,
                kafka=kafka,
                kafka_event_config=kafka_event_config,
            )
        )

    def handle_webhook_event(self, event: ClerkCompanyWebhookEvent):
        event_dict = {
            ClerkCompanyWebhookType.organization_created: self.create_company,
            ClerkCompanyWebhookType.organization_updated: self.update_company,
            ClerkCompanyWebhookType.organization_deleted: self.delete_company,
        }
        return event_dict[event.type](event)

    def _create_default_company_details(self, company_id: str):
        CompanyDetailsRepository(self.request_scope).set_sub_entity(
            data=SetCompanyDetails(),
            parent_collections={Entity.COMPANIES.value: company_id},
        )

    def create_company(self, event: ClerkCompanyWebhookEvent):
        create_event = ClerkCompany(**event.data)
        res = self.create(data=create_event, provided_id=create_event.id)
        self._create_default_company_details(company_id=create_event.id)
        return res

    def delete_company(self, event: ClerkCompanyWebhookEvent):
        delete_event = ClerkCompanyDeleted(**event.data)
        return self.delete(doc_id=delete_event.id)

    def update_company(self, event: ClerkCompanyWebhookEvent):
        update_event = ClerkCompany(**event.data)
        return self.update(doc_id=update_event.id, data=update_event)
